@@ -8,7 +8,7 @@ use hashbrown::{HashMap, HashSet};
 use crate::bytecode::InternalBytecode;
 use crate::error::{message_bad_eof, message_from_token, message_incorrect_requirement};
 use crate::expr::{Expr, ExprParser};
-use crate::lex::{Token, TokenStream, TokenType, Version};
+use crate::lex::{Token, TokenContext, TokenStream, TokenType, Version};
 
 /// The number of gates that are built in to the OpenQASM 2 language.  This is U and CX.
 const N_BUILTIN_GATES: usize = 2;
@@ -70,14 +70,35 @@ struct Condition {
     value: usize,
 }
 
+/// Find the first match for the partial [filename] in the directories along [path].  Returns
+/// `None` if the cannot be found.
+fn find_include_path(
+    filename: &std::path::Path,
+    path: &[std::path::PathBuf],
+) -> Option<std::path::PathBuf> {
+    for directory in path.iter() {
+        let mut absolute_path = directory.clone();
+        absolute_path.push(filename);
+        if absolute_path.is_file() {
+            return Some(absolute_path);
+        }
+    }
+    None
+}
+
 /// The state of the parser (but not its output).  This struct is opaque to the rest of the
 /// program; only its associated functions ever need to modify its internals.  The counts of
 /// qubits, clbits, classical registers and gates are necessary to efficiently assign index keys to
 /// new symbols as they arise.  We don't need to track quantum registers like this because no part
 /// of the output instruction set requires a reference to a quantum register, since we resolve any
 /// broadcast gate applications from within Rust.
-pub struct State<T: std::io::BufRead> {
-    tokens: TokenStream<T>,
+pub struct State {
+    tokens: Vec<TokenStream>,
+    /// The context object that owns all the text strings that back the tokens seen so far.  This
+    /// needs to be given as a read-only reference to the [Token] methods that extract information
+    /// based on the text they came from.
+    context: TokenContext,
+    include_path: Vec<std::path::PathBuf>,
     /// Mapping of name to global-scoped symbols.
     symbols: HashMap<String, GlobalSymbol>,
     /// Mapping of name to gate-scoped symbols.  This object only logically lasts for the duration
@@ -88,15 +109,17 @@ pub struct State<T: std::io::BufRead> {
     n_clbits: usize,
     n_cregs: usize,
     n_gates: usize,
-    /// Whether the parser has been called at all.
-    called: bool,
+    /// Whether a version statement is allowed in this position.
+    allow_version: bool,
 }
 
-impl<T: std::io::BufRead> State<T> {
+impl State {
     /// Create and initialise a state for the parser.
-    pub fn new(tokens: TokenStream<T>) -> Self {
+    pub fn new(tokens: TokenStream, include_path: Vec<std::path::PathBuf>) -> Self {
         let mut state = State {
-            tokens,
+            tokens: vec![tokens],
+            context: TokenContext::new(),
+            include_path,
             // For Qiskit-created circuits, all files will have the builtin gates and `qelib1.inc`,
             // so we allocate with that in mind.
             symbols: HashMap::with_capacity(N_BUILTIN_GATES + N_QELIB1_GATES),
@@ -105,7 +128,7 @@ impl<T: std::io::BufRead> State<T> {
             n_clbits: 0,
             n_cregs: 0,
             n_gates: 0,
-            called: false,
+            allow_version: true,
         };
         let dummy_token = Token::dummy();
         state.define_gate(&dummy_token, "U".into(), 3, 1).unwrap();
@@ -113,12 +136,47 @@ impl<T: std::io::BufRead> State<T> {
         state
     }
 
+    /// Get the next token available in the stack of token streams, popping and removing any
+    /// complete streams, except the base case.  Will only return `None` once all streams are
+    /// exhausted.
+    fn next_token(&mut self) -> Option<Token> {
+        let mut pointer = self.tokens.len() - 1;
+        while pointer > 0 {
+            let out = self.tokens[pointer].next(&mut self.context);
+            if out.is_some() {
+                return out;
+            }
+            self.tokens.pop();
+            pointer -= 1;
+        }
+        self.tokens[0].next(&mut self.context)
+    }
+
+    /// Peek the next token in the stack of token streams.  This does not remove any complete
+    /// streams yet.  Will only return `None` once all streams are exhausted.
+    fn peek_token(&mut self) -> Option<&Token> {
+        let mut pointer = self.tokens.len() - 1;
+        while pointer > 0 && self.tokens[pointer].peek(&mut self.context).is_none() {
+            pointer -= 1;
+        }
+        self.tokens[pointer].peek(&mut self.context)
+    }
+
+    /// Get the filename associated with the currently active token stream.
+    fn current_filename(&self) -> &std::ffi::OsStr {
+        &self.tokens[self.tokens.len() - 1].filename
+    }
+
     /// Take a token from the stream that is known to be present and correct, generally because it
     /// has already been peeked.  Panics if the token type is not correct.
     fn expect_known(&mut self, expected: TokenType) -> Token {
-        let out = self.tokens.next().unwrap();
+        let out = self.next_token().unwrap();
         if out.ttype != expected {
-            panic!()
+            panic!(
+                "expected '{}' but got '{}'",
+                expected.describe(),
+                out.ttype.describe()
+            )
         }
         out
     }
@@ -133,15 +191,15 @@ impl<T: std::io::BufRead> State<T> {
         required: &str,
         cause: &Token,
     ) -> Result<Token, String> {
-        let token = match self.tokens.next() {
-            None => return Err(message_bad_eof(&self.tokens.filename, required, cause)),
+        let token = match self.next_token() {
+            None => return Err(message_bad_eof(self.current_filename(), required, cause)),
             Some(token) => token,
         };
         if token.ttype == expected {
             Ok(token)
         } else {
             Err(message_incorrect_requirement(
-                &self.tokens.filename,
+                self.current_filename(),
                 required,
                 &token,
             ))
@@ -151,9 +209,9 @@ impl<T: std::io::BufRead> State<T> {
     /// Take the next token from the stream, if it is of the correct type.  Returns `None` and
     /// leaves the next token in the underlying iterator if it does not match.
     fn accept(&mut self, expected: TokenType) -> Option<Token> {
-        let peeked = self.tokens.peek();
+        let peeked = self.peek_token();
         if peeked.is_some() && peeked.unwrap().ttype == expected {
-            self.tokens.next()
+            self.next_token()
         } else {
             None
         }
@@ -161,7 +219,7 @@ impl<T: std::io::BufRead> State<T> {
 
     /// True if the next token in the stream matches the given type, and false if it doesn't.
     fn next_is(&mut self, expected: TokenType) -> bool {
-        let peeked = self.tokens.peek();
+        let peeked = self.peek_token();
         peeked.is_some() && peeked.unwrap().ttype == expected
     }
 
@@ -174,7 +232,7 @@ impl<T: std::io::BufRead> State<T> {
     fn accept_qarg(&mut self) -> Result<Option<Operand>, String> {
         let (name, name_token) = match self.accept(TokenType::Id) {
             None => return Ok(None),
-            Some(token) => (token.id(&self.tokens.context), token),
+            Some(token) => (token.id(&self.context), token),
         };
         let (register_size, register_start) = match self.symbols.get(&name) {
             Some(GlobalSymbol::Qreg { size, start }) => (*size, *start),
@@ -182,21 +240,21 @@ impl<T: std::io::BufRead> State<T> {
                 return Err(message_from_token(
                     &name_token,
                     &format!("'{}' is a classical register, not a quantum register", name),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             }
             Some(GlobalSymbol::Gate { .. }) => {
                 return Err(message_from_token(
                     &name_token,
                     &format!("'{}' is a gate, not a quantum register", name),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             }
             None => {
                 return Err(message_from_token(
                     &name_token,
                     &format!("'{}' is not defined in this scope", name),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             }
         };
@@ -209,14 +267,14 @@ impl<T: std::io::BufRead> State<T> {
     fn accept_qarg_gate(&mut self) -> Result<Option<Operand>, String> {
         let (name, name_token) = match self.accept(TokenType::Id) {
             None => return Ok(None),
-            Some(token) => (token.id(&self.tokens.context), token),
+            Some(token) => (token.id(&self.context), token),
         };
         match self.gate_symbols.get(&name) {
             Some(GateSymbol::Qubit { index }) => Ok(Some(Operand::Single(*index))),
             Some(GateSymbol::Parameter { .. }) => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is a parameter, not a qubit", name),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
             None => {
                 if self.symbols.contains_key(&name) {
@@ -228,13 +286,13 @@ impl<T: std::io::BufRead> State<T> {
                     Err(message_from_token(
                         &name_token,
                         &format!("'{}' is a {}, not a qubit", name, bad_type),
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ))
                 } else {
                     Err(message_from_token(
                         &name_token,
                         &format!("'{}' is not defined in this scope", name),
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ))
                 }
             }
@@ -244,18 +302,18 @@ impl<T: std::io::BufRead> State<T> {
     /// Take a complete quantum argument from the token stream, returning an error message if one
     /// is not present.
     fn require_qarg(&mut self, instruction: &Token) -> Result<Operand, String> {
-        match self.tokens.peek().map(|tok| tok.ttype) {
+        match self.peek_token().map(|tok| tok.ttype) {
             Some(TokenType::Id) => self.accept_qarg().map(Option::unwrap),
             Some(_) => {
-                let token = self.tokens.next();
+                let token = self.next_token();
                 Err(message_incorrect_requirement(
-                    &self.tokens.filename,
+                    self.current_filename(),
                     "a quantum argument",
                     &token.unwrap(),
                 ))
             }
             None => Err(message_bad_eof(
-                &self.tokens.filename,
+                self.current_filename(),
                 "a quantum argument",
                 instruction,
             )),
@@ -272,7 +330,7 @@ impl<T: std::io::BufRead> State<T> {
     fn accept_carg(&mut self) -> Result<Option<Operand>, String> {
         let (name, name_token) = match self.accept(TokenType::Id) {
             None => return Ok(None),
-            Some(token) => (token.id(&self.tokens.context), token),
+            Some(token) => (token.id(&self.context), token),
         };
         let (register_size, register_start) = match self.symbols.get(&name) {
             Some(GlobalSymbol::Creg { size, start, .. }) => (*size, *start),
@@ -280,21 +338,21 @@ impl<T: std::io::BufRead> State<T> {
                 return Err(message_from_token(
                     &name_token,
                     &format!("'{}' is a quantum register, not a classical one", name),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             }
             Some(GlobalSymbol::Gate { .. }) => {
                 return Err(message_from_token(
                     &name_token,
                     &format!("'{}' is a gate, not a classical register", name),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             }
             None => {
                 return Err(message_from_token(
                     &name_token,
                     &format!("'{}' is not defined in this scope", name),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             }
         };
@@ -305,18 +363,18 @@ impl<T: std::io::BufRead> State<T> {
     /// Take a complete classical argument from the token stream, returning an error message if one
     /// is not present.
     fn require_carg(&mut self, instruction: &Token) -> Result<Operand, String> {
-        match self.tokens.peek().map(|tok| tok.ttype) {
+        match self.peek_token().map(|tok| tok.ttype) {
             Some(TokenType::Id) => self.accept_carg().map(Option::unwrap),
             Some(_) => {
-                let token = self.tokens.next();
+                let token = self.next_token();
                 Err(message_incorrect_requirement(
-                    &self.tokens.filename,
+                    self.current_filename(),
                     "a classical argument",
                     &token.unwrap(),
                 ))
             }
             None => Err(message_bad_eof(
-                &self.tokens.filename,
+                self.current_filename(),
                 "a classical argument",
                 instruction,
             )),
@@ -337,7 +395,7 @@ impl<T: std::io::BufRead> State<T> {
             None => return Ok(Operand::Range(register_size, register_start)),
         };
         let index_token = self.expect(TokenType::Integer, "an integer index", &lbracket_token)?;
-        let index = index_token.int(&self.tokens.context);
+        let index = index_token.int(&self.context);
         self.expect(TokenType::RBracket, "a closing bracket", &lbracket_token)?;
         if index < register_size {
             Ok(Operand::Single(register_start + index))
@@ -348,7 +406,7 @@ impl<T: std::io::BufRead> State<T> {
                     "index {} is out-of-range for register '{}' of size {}",
                     index, name, register_size
                 ),
-                &self.tokens.filename,
+                self.current_filename(),
             ))
         }
     }
@@ -361,7 +419,7 @@ impl<T: std::io::BufRead> State<T> {
     fn parse_version(&mut self) -> Result<usize, String> {
         let openqasm_token = self.expect_known(TokenType::OpenQASM);
         let version_token = self.expect(TokenType::Version, "version number", &openqasm_token)?;
-        match version_token.version(&self.tokens.context) {
+        match version_token.version(&self.context) {
             Version {
                 major: 2,
                 minor: Some(0) | None,
@@ -370,9 +428,9 @@ impl<T: std::io::BufRead> State<T> {
                 &version_token,
                 &format!(
                     "can only handle OpenQASM 2.0, but given {}",
-                    version_token.text(&self.tokens.context),
+                    version_token.text(&self.context),
                 ),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
         }?;
         self.expect(TokenType::Semicolon, ";", &openqasm_token)?;
@@ -389,12 +447,12 @@ impl<T: std::io::BufRead> State<T> {
     ) -> Result<usize, String> {
         let gate_token = self.expect_known(TokenType::Gate);
         let name_token = self.expect(TokenType::Id, "an identifier", &gate_token)?;
-        let name = name_token.id(&self.tokens.context);
+        let name = name_token.id(&self.context);
         // Parse the gate parameters (if any) into the symbol take.
         let mut n_params = 0usize;
         if let Some(lparen_token) = self.accept(TokenType::LParen) {
             while let Some(param_token) = self.accept(TokenType::Id) {
-                let param_name = param_token.id(&self.tokens.context);
+                let param_name = param_token.id(&self.context);
                 match self.gate_symbols.insert(
                     param_name.to_owned(),
                     GateSymbol::Parameter { index: n_params },
@@ -403,14 +461,14 @@ impl<T: std::io::BufRead> State<T> {
                         return Err(message_from_token(
                             &param_token,
                             &format!("'{}' is already defined as a parameter", param_name),
-                            &self.tokens.filename,
+                            self.current_filename(),
                         ))
                     }
                     Some(GateSymbol::Qubit { .. }) => {
                         return Err(message_from_token(
                             &param_token,
                             &format!("'{}' is already defined as a qubit", param_name),
-                            &self.tokens.filename,
+                            self.current_filename(),
                         ))
                     }
                     None => (),
@@ -425,7 +483,7 @@ impl<T: std::io::BufRead> State<T> {
         // Parse the quantum parameters into the symbol table.
         let mut n_qubits = 0usize;
         while let Some(qubit_token) = self.accept(TokenType::Id) {
-            let qubit_name = qubit_token.id(&self.tokens.context).to_owned();
+            let qubit_name = qubit_token.id(&self.context).to_owned();
             match self
                 .gate_symbols
                 .insert(qubit_name.to_owned(), GateSymbol::Qubit { index: n_qubits })
@@ -434,14 +492,14 @@ impl<T: std::io::BufRead> State<T> {
                     return Err(message_from_token(
                         &qubit_token,
                         &format!("'{}' is already defined as a parameter", qubit_name),
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ))
                 }
                 Some(GateSymbol::Qubit { .. }) => {
                     return Err(message_from_token(
                         &qubit_token,
                         &format!("'{}' is already defined as a qubit", qubit_name),
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ))
                 }
                 None => (),
@@ -452,9 +510,9 @@ impl<T: std::io::BufRead> State<T> {
             }
         }
         if n_qubits == 0 {
-            return if self.tokens.peek().is_none() {
+            return if self.peek_token().is_none() {
                 Err(message_bad_eof(
-                    &self.tokens.filename,
+                    self.current_filename(),
                     "a qubit identifier",
                     &gate_token,
                 ))
@@ -462,7 +520,7 @@ impl<T: std::io::BufRead> State<T> {
                 Err(message_from_token(
                     &gate_token,
                     "gates must act on at least one qubit",
-                    &self.tokens.filename,
+                    self.current_filename(),
                 ))
             };
         }
@@ -475,7 +533,7 @@ impl<T: std::io::BufRead> State<T> {
         // to do the right thing.
         let mut statements = 0usize;
         loop {
-            match self.tokens.peek().map(|tok| tok.ttype) {
+            match self.peek_token().map(|tok| tok.ttype) {
                 Some(TokenType::Id) => statements += self.parse_gate_application(bc, None, true)?,
                 Some(TokenType::Barrier) => statements += self.parse_barrier(bc, Some(n_qubits))?,
                 Some(TokenType::RBrace) => {
@@ -483,19 +541,19 @@ impl<T: std::io::BufRead> State<T> {
                     break;
                 }
                 Some(_) => {
-                    let token = self.tokens.next().unwrap();
+                    let token = self.next_token().unwrap();
                     return Err(message_from_token(
                         &token,
                         &format!(
                             "only gate applications are valid within a 'gate' body, but saw {}",
-                            token.text(&self.tokens.context)
+                            token.text(&self.context)
                         ),
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ));
                 }
                 None => {
                     return Err(message_bad_eof(
-                        &self.tokens.filename,
+                        self.current_filename(),
                         "a closing brace '}' of the gate body",
                         &lbrace_token,
                     ))
@@ -517,7 +575,7 @@ impl<T: std::io::BufRead> State<T> {
         let opaque_token = self.expect_known(TokenType::Opaque);
         let name = self
             .expect(TokenType::Id, "an identifier", &opaque_token)?
-            .text(&self.tokens.context)
+            .text(&self.context)
             .to_owned();
         let mut n_params = 0usize;
         if let Some(lparen_token) = self.accept(TokenType::LParen) {
@@ -541,7 +599,7 @@ impl<T: std::io::BufRead> State<T> {
             return Err(message_from_token(
                 &opaque_token,
                 "gates must act on at least one qubit",
-                &self.tokens.filename,
+                self.current_filename(),
             ));
         }
         bc.push(Some(InternalBytecode::DeclareOpaque {
@@ -563,7 +621,7 @@ impl<T: std::io::BufRead> State<T> {
         in_gate: bool,
     ) -> Result<usize, String> {
         let name_token = self.expect_known(TokenType::Id);
-        let name = name_token.id(&self.tokens.context);
+        let name = name_token.id(&self.context);
         let (index, n_params, n_qubits) = match self.symbols.get(&name) {
             Some(GlobalSymbol::Gate {
                 n_params,
@@ -581,12 +639,12 @@ impl<T: std::io::BufRead> State<T> {
                         "quantum"
                     }
                 ),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
             None => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is not defined in this scope", name),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
         }?;
         let parameters = self.expect_gate_parameters(&name_token, n_params, in_gate)?;
@@ -607,7 +665,7 @@ impl<T: std::io::BufRead> State<T> {
             }
         }
         if qargs.len() != n_qubits {
-            return match self.tokens.peek().map(|tok| tok.ttype) {
+            return match self.peek_token().map(|tok| tok.ttype) {
                 Some(TokenType::Semicolon) => Err(message_from_token(
                     &name_token,
                     &format!(
@@ -617,15 +675,15 @@ impl<T: std::io::BufRead> State<T> {
                         if n_qubits == 1 { "" } else { "s" },
                         qargs.len()
                     ),
-                    &self.tokens.filename,
+                    self.current_filename(),
                 )),
                 Some(_) => Err(message_incorrect_requirement(
-                    &self.tokens.filename,
+                    self.current_filename(),
                     "the end of the argument list",
                     &name_token,
                 )),
                 None => Err(message_bad_eof(
-                    &self.tokens.filename,
+                    self.current_filename(),
                     "the end of the argument list",
                     &name_token,
                 )),
@@ -661,6 +719,7 @@ impl<T: std::io::BufRead> State<T> {
             while !self.next_is(TokenType::RParen) {
                 let mut expr_parser = ExprParser {
                     tokens: &mut self.tokens,
+                    context: &mut self.context,
                     gate_symbols: &self.gate_symbols,
                 };
                 parameters.push(expr_parser.parse_expression(&lparen_token)?);
@@ -676,6 +735,7 @@ impl<T: std::io::BufRead> State<T> {
             while !self.next_is(TokenType::RParen) {
                 let mut expr_parser = ExprParser {
                     tokens: &mut self.tokens,
+                    context: &mut self.context,
                     gate_symbols: &self.gate_symbols,
                 };
                 match expr_parser.parse_expression(&lparen_token)? {
@@ -684,7 +744,7 @@ impl<T: std::io::BufRead> State<T> {
                         return Err(message_from_token(
                             &lparen_token,
                             "non-constant expression in program body",
-                            &self.tokens.filename,
+                            self.current_filename(),
                         ))
                     }
                 }
@@ -701,12 +761,12 @@ impl<T: std::io::BufRead> State<T> {
                 name_token,
                 &format!(
                     "'{}' takes {} parameter{}, but got {}",
-                    &name_token.text(&self.tokens.context),
+                    &name_token.text(&self.context),
                     n_params,
                     if n_params == 1 { "" } else { "s" },
                     seen_params
                 ),
-                &self.tokens.filename,
+                self.current_filename(),
             ));
         }
         Ok(parameters)
@@ -731,7 +791,7 @@ impl<T: std::io::BufRead> State<T> {
                     return Err(message_from_token(
                         instruction,
                         "duplicate qubits in gate application",
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ));
                 }
                 Some(vec![*left, *right])
@@ -759,7 +819,7 @@ impl<T: std::io::BufRead> State<T> {
                         return Err(message_from_token(
                             instruction,
                             "duplicate qubits in gate application",
-                            &self.tokens.filename,
+                            self.current_filename(),
                         ));
                     }
                 }
@@ -768,7 +828,7 @@ impl<T: std::io::BufRead> State<T> {
                         return Err(message_from_token(
                             instruction,
                             "cannot resolve broadcast in gate application",
-                            &self.tokens.filename,
+                            self.current_filename(),
                         ));
                     }
                     for index in *start..*start + *size {
@@ -776,7 +836,7 @@ impl<T: std::io::BufRead> State<T> {
                             return Err(message_from_token(
                                 instruction,
                                 "duplicate qubits in gate application",
-                                &self.tokens.filename,
+                                self.current_filename(),
                             ));
                         }
                     }
@@ -892,42 +952,42 @@ impl<T: std::io::BufRead> State<T> {
         self.expect(TokenType::Equals, "'=='", &if_token)?;
         let value = self
             .expect(TokenType::Integer, "an integer", &if_token)?
-            .int(&self.tokens.context);
+            .int(&self.context);
         self.expect(TokenType::RParen, "')'", &lparen_token)?;
-        let name = name_token.id(&self.tokens.context);
+        let name = name_token.id(&self.context);
         let creg = match self.symbols.get(&name) {
             Some(GlobalSymbol::Creg { index, .. }) => Ok(*index),
             Some(GlobalSymbol::Qreg { .. }) => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is a quantum register, not a classical register", name),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
             Some(GlobalSymbol::Gate { .. }) => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is a gate, not a classical register", name),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
             None => Err(message_from_token(
                 &name_token,
                 &format!("'{}' is not defined in this scope", name),
-                &self.tokens.filename,
+                self.current_filename(),
             )),
         }?;
         let condition = Some(Condition { creg, value });
-        match self.tokens.peek().map(|tok| tok.ttype) {
+        match self.peek_token().map(|tok| tok.ttype) {
             Some(TokenType::Id) => self.parse_gate_application(bc, condition, false),
             Some(TokenType::Measure) => self.parse_measure(bc, condition),
             Some(TokenType::Reset) => self.parse_reset(bc, condition),
             Some(_) => {
-                let token = self.tokens.next();
+                let token = self.next_token();
                 Err(message_incorrect_requirement(
-                    &self.tokens.filename,
+                    self.current_filename(),
                     "a gate application, measurement or reset",
                     &token.unwrap(),
                 ))
             }
             None => Err(message_bad_eof(
-                &self.tokens.filename,
+                self.current_filename(),
                 "a gate, measurement or reset to condition",
                 &if_token,
             )),
@@ -1021,7 +1081,7 @@ impl<T: std::io::BufRead> State<T> {
                 _ => Err(message_from_token(
                     &measure_token,
                     "cannot resolve broadcast in measurement",
-                    &self.tokens.filename,
+                    self.current_filename(),
                 )),
             }
         } else {
@@ -1044,7 +1104,7 @@ impl<T: std::io::BufRead> State<T> {
                 _ => Err(message_from_token(
                     &measure_token,
                     "cannot resolve broadcast in measurement",
-                    &self.tokens.filename,
+                    self.current_filename(),
                 )),
             }
         }
@@ -1100,11 +1160,11 @@ impl<T: std::io::BufRead> State<T> {
     fn parse_creg(&mut self, bc: &mut Vec<Option<InternalBytecode>>) -> Result<usize, String> {
         let creg_token = self.expect_known(TokenType::Creg);
         let name_token = self.expect(TokenType::Id, "a valid identifier", &creg_token)?;
-        let name = name_token.id(&self.tokens.context);
+        let name = name_token.id(&self.context);
         let lbracket_token = self.expect(TokenType::LBracket, "'['", &creg_token)?;
         let size = self
             .expect(TokenType::Integer, "an integer", &lbracket_token)?
-            .int(&self.tokens.context);
+            .int(&self.context);
         self.expect(TokenType::RBracket, "']'", &lbracket_token)?;
         self.expect(TokenType::Semicolon, "';'", &creg_token)?;
         bc.push(Some(InternalBytecode::DeclareCreg {
@@ -1125,11 +1185,8 @@ impl<T: std::io::BufRead> State<T> {
         {
             return Err(message_from_token(
                 &name_token,
-                &format!(
-                    "'{}' is already defined",
-                    name_token.id(&self.tokens.context)
-                ),
-                &self.tokens.filename,
+                &format!("'{}' is already defined", name_token.id(&self.context)),
+                self.current_filename(),
             ));
         }
         self.n_clbits += size;
@@ -1143,11 +1200,11 @@ impl<T: std::io::BufRead> State<T> {
     fn parse_qreg(&mut self, bc: &mut Vec<Option<InternalBytecode>>) -> Result<usize, String> {
         let qreg_token = self.expect_known(TokenType::Qreg);
         let name_token = self.expect(TokenType::Id, "a valid identifier", &qreg_token)?;
-        let name = name_token.id(&self.tokens.context);
+        let name = name_token.id(&self.context);
         let lbracket_token = self.expect(TokenType::LBracket, "'['", &qreg_token)?;
         let size = self
             .expect(TokenType::Integer, "an integer", &lbracket_token)?
-            .int(&self.tokens.context);
+            .int(&self.context);
         self.expect(TokenType::RBracket, "']'", &lbracket_token)?;
         self.expect(TokenType::Semicolon, "';'", &qreg_token)?;
         bc.push(Some(InternalBytecode::DeclareQreg {
@@ -1167,11 +1224,8 @@ impl<T: std::io::BufRead> State<T> {
         {
             return Err(message_from_token(
                 &name_token,
-                &format!(
-                    "'{}' is already defined",
-                    name_token.id(&self.tokens.context)
-                ),
-                &self.tokens.filename,
+                &format!("'{}' is already defined", name_token.id(&self.context)),
+                self.current_filename(),
             ));
         }
         self.n_qubits += size;
@@ -1188,17 +1242,34 @@ impl<T: std::io::BufRead> State<T> {
         let filename_token =
             self.expect(TokenType::Filename, "a filename string", &include_token)?;
         self.expect(TokenType::Semicolon, "';'", &include_token)?;
-        let filename = filename_token.filename(&self.tokens.context);
+        let filename = filename_token.filename(&self.context);
         if filename == "qelib1.inc" {
             self.include_qelib1(&include_token)?;
             bc.push(Some(InternalBytecode::SpecialInclude { name: filename }));
             Ok(1)
         } else {
-            Err(message_from_token(
-                &filename_token,
-                &format!("can only currently handle 'qelib1.inc', not '{}'", filename),
-                &self.tokens.filename,
-            ))
+            let base_filename = std::path::PathBuf::from(&filename);
+            let absolute_filename = find_include_path(&base_filename, &self.include_path)
+                .ok_or_else(|| {
+                    message_from_token(
+                        &filename_token,
+                        &format!(
+                            "unable to find '{}' in the include search path",
+                            base_filename.display()
+                        ),
+                        self.current_filename(),
+                    )
+                })?;
+            let new_stream = TokenStream::from_path(absolute_filename).map_err(|err| {
+                message_from_token(
+                    &filename_token,
+                    &format!("unable to open file '{}' for reading: {}", &filename, err),
+                    self.current_filename(),
+                )
+            })?;
+            self.tokens.push(new_stream);
+            self.allow_version = true;
+            Ok(0)
         }
     }
 
@@ -1247,7 +1318,7 @@ impl<T: std::io::BufRead> State<T> {
             Err(message_from_token(
                 owner,
                 &format!("name '{}' is already defined", name),
-                &self.tokens.filename,
+                self.current_filename(),
             ))
         } else {
             self.symbols.insert(
@@ -1275,9 +1346,9 @@ impl<T: std::io::BufRead> State<T> {
         &mut self,
         bc: &mut Vec<Option<InternalBytecode>>,
     ) -> Result<Option<usize>, String> {
-        let called = self.called;
-        self.called = true;
-        if let Some(ttype) = self.tokens.peek().map(|tok| tok.ttype) {
+        let allow_version = self.allow_version;
+        self.allow_version = false;
+        if let Some(ttype) = self.peek_token().map(|tok| tok.ttype) {
             match match ttype {
                 TokenType::Id => self.parse_gate_application(bc, None, false),
                 TokenType::Creg => self.parse_creg(bc),
@@ -1290,26 +1361,26 @@ impl<T: std::io::BufRead> State<T> {
                 TokenType::Opaque => self.parse_opaque_definition(bc),
                 TokenType::Gate => self.parse_gate_definition(bc),
                 TokenType::OpenQASM => {
-                    if called {
-                        let token = self.tokens.next().unwrap();
+                    if allow_version {
+                        self.parse_version()
+                    } else {
+                        let token = self.next_token().unwrap();
                         Err(message_from_token(
                             &token,
                             "only the first statement may be a version declaration",
-                            &self.tokens.filename,
+                            self.current_filename(),
                         ))
-                    } else {
-                        self.parse_version()
                     }
                 }
                 _ => {
-                    let token = self.tokens.next().unwrap();
+                    let token = self.next_token().unwrap();
                     Err(message_from_token(
                         &token,
                         &format!(
                             "needed a start-of-statement token, but instead got {}",
-                            token.text(&self.tokens.context)
+                            token.text(&self.context)
                         ),
-                        &self.tokens.filename,
+                        self.current_filename(),
                     ))
                 }
             }? {
